# coding=utf8

import json

from context import get_json_value
from util import dumpjs, is_string_set, print_traceback
from datamodel import is_objecttype_valid, get_field_info
from tag import TagFilter

cloudsight_api_languages = {
    'ar': 'ar',
    'cs-CZ': 'cs',
    'de-DE': 'de',
    'en-US': 'en',
    'es-ES': 'es',
    'fa': 'fa',
    'fr-FR': 'fr',
    'it-IT': 'it',
    'ja-Jpan': 'ja',
    'ka-GE': 'ka',
    'ko-Kore': 'ko',
    'nl-NL': 'nl',
    'pl-PL': 'pl',
    'ru-RU': 'ru',
    'zh-Hans': 'zh'
}


class CloudsightConfigElement(object):
    def __init__(self,
                 debug_name,
                 api_key,
                 objecttype,
                 asset_field,
                 asset_version,
                 subject_field,
                 keyword_field_map,
                 timestamp_field,
                 tagfilter,
                 language,
                 min_age_days,
                 enabled):

        self.debug_name = debug_name
        self.api_key = api_key
        self.objecttype = objecttype
        self.asset_field = asset_field
        self.asset_version = asset_version
        self.subject_field = subject_field
        self.keyword_field_map = keyword_field_map
        self.timestamp_field = timestamp_field
        self.tagfilter = tagfilter
        self.language = language
        self.min_age_days = min_age_days
        self.enabled = enabled

    def to_dict(self):
        d = {
            'debug_name': self.debug_name,
            'api_key': self.api_key,
            'objecttype': self.objecttype,
            'asset_field': self.asset_field.to_dict(),
            'asset_version': self.asset_version,
            'subject_field': self.subject_field.to_dict() if self.subject_field is not None else None,
            'keyword_field_map': {},
            'timestamp_field': self.timestamp_field.to_dict(),
            'tagfilter': self.tagfilter.to_dict(),
            'language': self.language,
            'min_age_days': self.min_age_days,
            'enabled': self.enabled
        }

        for k in self.keyword_field_map:
            d['keyword_field_map'][k] = self.keyword_field_map[k].to_dict()

        return d


class CloudsightConfig(object):
    def __init__(self,
                 easydb_user,
                 easydb_password,
                 easydb_session_token,
                 cloudsight_config_list,
                 enabled,
                 start_now,
                 request_status_repititions,
                 request_status_delay):

        self.easydb_user = easydb_user
        self.easydb_password = easydb_password
        self.easydb_session_token = easydb_session_token
        self.cloudsight_config_list = cloudsight_config_list
        self.enabled = enabled
        self.start_now = start_now
        self.request_status_repititions = request_status_repititions
        self.request_status_delay = request_status_delay

    def __str__(self):
        d = {
            'enabled': self.enabled,
            'start_now': self.start_now,
            'request_status_repititions': self.request_status_repititions,
            'request_status_delay': self.request_status_delay,
            'cloudsight_config_list': [c.to_dict() for c in self.cloudsight_config_list]
        }
        if self.easydb_user is not None:
            d['easydb_user'] = self.easydb_user
        if self.easydb_session_token is not None:
            d['easydb_session_token'] = self.easydb_session_token

        return dumpjs(d)


def parse_cloudsight_from_baseconfig(easydb_context, config_def):

    logger = easydb_context.get_logger(
        'pf.plugin.base.auto_keyworder.cloudsight.config')
    datamodel = easydb_context.get_datamodel()
    configs = []

    cl = get_json_value(config_def, 'configs')
    i = 0
    if isinstance(cl, list):
        for c in cl:
            try:
                # read the service first and ignore all configurations that are not for Cloudsight.ai
                _service = get_json_value(c, 'service')
                if _service != 'Cloudsight.ai':
                    logger.debug('[%d] skip irrelevant service %s' %
                                 (i, str(_service)))
                    continue

                _name = get_json_value(c, 'name')
                if not is_string_set(_name):
                    _name = '<unnamed>'

                _enabled = get_json_value(c, 'config_enabled')
                if not isinstance(_enabled, bool):
                    _enabled = False
                if not _enabled:
                    logger.debug(
                        '[%d \'%s\'] not enabled -> skip config entry' % (i, _name))
                    continue

                _api_key = get_json_value(c, 'api_key')
                if not is_string_set(_api_key):
                    logger.debug(
                        '[%d \'%s\'] api_key invalid/not set -> skip config entry' % (i, _name))
                    continue

                # check if the objecttype is valid
                _objecttype = get_json_value(c, 'objecttype')
                if not is_string_set(_objecttype):
                    logger.debug(
                        '[%d \'%s\'] objecttype invalid/not set -> skip config entry' % (i, _name))
                    continue
                is_objecttype_valid(datamodel, _objecttype)

                _asset_field = get_json_value(c, 'asset_field')
                if not is_string_set(_asset_field):
                    logger.debug(
                        '[%d \'%s\'] asset_field invalid/not set -> skip config entry' % (i, _name))
                    continue
                _asset_field = get_field_info(
                    datamodel, _objecttype, _asset_field, ['eas'])

                _asset_version = get_json_value(c, 'asset_version')
                if not is_string_set(_asset_version):
                    logger.debug(
                        '[%d \'%s\'] asset_version invalid/not set -> skip config entry' % (i, _name))
                    continue

                _subject_field = get_json_value(c, 'subject_field')
                if _subject_field is not None:
                    if not is_string_set(_subject_field):
                        _subject_field = None
                    else:
                        _subject_field = get_field_info(datamodel, _objecttype, _subject_field, [
                            'text',
                            'text_l10n',
                            'text_oneline',
                            'text_l10n_oneline',
                            'string'
                        ])

                # process the keyword target fields:
                # - remove objecttype prefix
                # - find field in datamodel
                # - determine if the field is inside a nested table or on top level or in a linked object
                _keyword_map = {}
                valid = False
                for k in [
                    'keyword_categories_field',
                    'keyword_quantity_field',
                    'keyword_gender_field',
                    'keyword_material_field',
                    'keyword_color_field'
                ]:
                    _keyword_field = get_json_value(c, k)
                    if not is_string_set(_keyword_field):
                        continue
                    _keyword_map[k] = get_field_info(datamodel, _objecttype, _keyword_field, [
                        'text',
                        'text_l10n',
                        'text_oneline',
                        'text_l10n_oneline',
                        'string'
                    ], True)
                    valid = True
                if not valid:
                    logger.debug(
                        '[%d \'%s\'] none of the keyword fields is set -> skip config entry' % (i, _name))

                if _subject_field is not None:
                    valid = True
                    for k in _keyword_map:
                        if _subject_field.equals(_keyword_map[k]):
                            logger.debug(
                                '[%d \'%s\'] subject_field must not be the same as %s -> skip config entry' % (i, _name, k))
                            valid = False
                            break
                    if not valid:
                        continue

                _timestamp_field = get_json_value(c, 'timestamp_field')
                if not is_string_set(_timestamp_field):
                    logger.debug(
                        '[%d \'%s\'] timestamp_field invalid/not set -> skip config entry' % (i, _name))
                    continue
                _timestamp_field = get_field_info(
                    datamodel, _objecttype, _timestamp_field, ['datetime'])

                _language = get_json_value(c, 'language')
                if not is_string_set(_language) or _language not in cloudsight_api_languages:
                    logger.debug(
                        '[%d \'%s\'] language invalid/not set -> skip config entry' % (i, _name))
                    continue

                _tagfilter_def = get_json_value(c, 'tagfilter')
                if not isinstance(_tagfilter_def, dict):
                    logger.debug(
                        '[%d \'%s\'] tagfilter invalid/not set -> skip config entry' % (i, _name))
                    continue
                _tagfilter = TagFilter(_tagfilter_def)

                _max_age = get_json_value(c, 'min_age_days')
                if not isinstance(_max_age, int) or _max_age < 1:
                    logger.debug(
                        '[%d \'%s\'] min_age_days invalid/not set -> use default value 7 for config' % (i, _name))
                    _max_age = 7

                configs.append(
                    CloudsightConfigElement(_name,
                                            _api_key,
                                            _objecttype,
                                            _asset_field,
                                            _asset_version,
                                            _subject_field,
                                            _keyword_map,
                                            _timestamp_field,
                                            _tagfilter,
                                            _language,
                                            _max_age,
                                            _enabled))
            except Exception as e:
                print_traceback(logger, e)
                logger.debug('[%d \'%s\'] skip config: %s' %
                             (i, _name, e.message))
            i += 1

    _enabled = get_json_value(config_def, 'enabled')
    if not isinstance(_enabled, bool):
        logger.debug('enabled invalid/not set: use default value False')
        _enabled = False

    _start_now = get_json_value(config_def, 'start_now')
    if not isinstance(_start_now, bool):
        logger.debug('start_now invalid/not set: use default value False')
        _start_now = False

    _easydb_session_token = get_json_value(
        config_def, 'easydb_session_token')
    _easydb_user = None
    _easydb_password = None
    # if the session token is set, ignore user and password
    if not is_string_set(_easydb_session_token):
        _easydb_user = get_json_value(config_def, 'easydb_user')
        if not is_string_set(_easydb_user):
            if _enabled:
                logger.debug('easydb_user invalid/not set')
            _easydb_user = None

        _easydb_password = get_json_value(config_def, 'easydb_password')
        if not is_string_set(_easydb_password):
            if _enabled:
                logger.debug('easydb_password invalid/not set')
            _easydb_password = None

        _easydb_session_token = None

    _request_status_repititions = get_json_value(
        config_def, 'request_status_repititions')
    if not isinstance(_request_status_repititions, int) or _request_status_repititions < 1:
        _request_status_repititions = 5
        logger.debug('request_status_repititions invalid/not set: use default value %' %
                     _request_status_repititions)

    _request_status_delay = get_json_value(config_def, 'request_status_delay')
    if _request_status_delay < 1:
        _request_status_delay = 5
        logger.debug(
            'request_status_delay invalid/not set: use default value %d' % _request_status_delay)

    config = CloudsightConfig(_easydb_user,
                              _easydb_password,
                              _easydb_session_token,
                              configs,
                              _enabled,
                              _start_now,
                              _request_status_repititions,
                              _request_status_delay)
    logger.debug('config: %s' % config)
    return config
